/**********************************
 * blog.iriscores.com 2016/2022
 * Author: vojta.ters@iriscores.com
 **********************************/

// ------------------------------
// -------- Cuda Includes -------
#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <cufft.h>
#include <math_constants.h>

// -----------------------------
// ------- C++ Includes --------
#include <stdio.h>
#include <math.h>
#include <string>
#include <iostream>

// -----------------------------
// ------ Custom Includes ------
#include "Kernels.h"

// --------------------------------
// -------- Matlab Includes -------
#include <matrix.h>
#include <mex.h>

#define HANDLE_CUDA( err )  (Handle_Error( err, __FILE__, __LINE__ ))
#define RADIX_2 2
#define RADIX_4 4
#define RADIX_8 8

/******************************************************
 *@brief Error Handling Function of all the cuda Calls
 ****************************************************/
void Handle_Error(cudaError_t err, const char* file, int line) {
	if (err != cudaSuccess) {
		std::string Message = std::string(cudaGetErrorString(err)) + std::string(" in ") + std::string(file) + std::string(" at line: ") + std::to_string(line) + std::string("\n");
		mexErrMsgTxt(Message.c_str());
	}
}

/************************************************
 *@brief: Calculates Logarithm with a custom Base
 ************************************************/
template<typename Type>
Type BaseLog(Type num, Type Base) {
	return static_cast<Type>(log(static_cast<double>(num)) / log(static_cast<double>(Base)));
}


// ----------------------------
// ------ Global Arrays -------
struct DeviceArrays {
	int NumDevices;
	cudaDeviceProp Prop;
	cudaEvent_t Start;
	cudaEvent_t Stop;
	cufftComplex* Data1;
	cufftComplex* Data2;
	cufftComplex* Twiddles;
	int* Indexes;
	dim3 THREADS;
	dim3 BLOCKS;
}Dev;

/**************************************************
 *@brief Main Matlab Function
 *@param nlhs - Amount of LeftHandSide Parameters
 *@param nrhs - Amount of RightHandside Parameters
 *@param plhs - Array of Output Arguments
 *@param prhs - Array of Input Arguments
 ***************************************************/
void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[]){

	// ----------------------------------------
	// ------- Check Amount of Aguments -------
	if (nrhs < 2) { mexErrMsgTxt("Not Enough Input Arguments!\n"); }

	const mxArray* matlabInputData = prhs[0];
	const mxArray* matlabInputAdd = prhs[1];
	int Base;

	// --------------------------------
	// ------ Check Class Types -------
	mxClassID ClOptions = mxGetClassID(matlabInputAdd);
	mxClassID ClData    = mxGetClassID(matlabInputData);
	if (ClOptions != mxSINGLE_CLASS || ClData != mxSINGLE_CLASS) {
		if (ClOptions != mxSINGLE_CLASS) { std::string Message = std::string("Options should be in SINGLE precision format!\n"); mexErrMsgTxt(Message.c_str()); }
		if (ClData != mxSINGLE_CLASS) { std::string Message = std::string("Data	should be in SINGLE precision format!\n"); mexErrMsgTxt(Message.c_str()); }
		return;
	}
	
	// --------------------------------------------
	// ------- Check for Correct Complexity -------
	if (!mxIsComplex(matlabInputData)){
		std::string Message = std::string("Error Invalid Complexity!\n");
		mexErrMsgTxt(Message.c_str());
	}

	// ---------------------------------------------------
	// ------- GET CUDA RUNTIME AND DRIVER VERSION -------
	int RuntimeVersion = 0;
	int DriverVersion = 0;
	int NumDevices = 0;
	cudaRuntimeGetVersion(&RuntimeVersion);
	HANDLE_CUDA(cudaGetLastError());
	cudaDriverGetVersion(&DriverVersion);
	HANDLE_CUDA(cudaGetLastError());

	// ------------------------------------
    // ------- CHECK DRIVER SUPPORT -------
	if (RuntimeVersion > DriverVersion) {
		std::string Message = std::string("Cuda Runtime Version : ") + std::to_string(RuntimeVersion) + std::string("\n");
		Message += std::string("Cuda Driver Version  : ") + std::to_string(DriverVersion) + std::string("\n");
		Message += std::string("Internal Error. Insufficient Driver, please update Graphics Drivers to the latest version\n");
		mexErrMsgTxt(Message.c_str());
	}

	// -------------------------------------------
	// ------- CHECK FOR AMOUNT OF DEVICES -------
	cudaError_t Err = cudaGetDeviceCount(&NumDevices);
	if (Err == cudaErrorNoDevice) {
		std::string Message = std::string("Internal Error. No CUDA Capable Device Found!\n");
		mexErrMsgTxt(Message.c_str());
	}

	// ------------------------------------------
	// ------ Figure Out Input Dimensions -------
	int Nx = static_cast<int>(mxGetNumberOfElements(matlabInputData));
	int Na = static_cast<int>(mxGetNumberOfElements(matlabInputAdd));

	mxSingle* Options = mxGetSingles(matlabInputAdd);
	mxComplexSingle* MatData_In = mxGetComplexSingles(matlabInputData);

	// ----------------------------------------
	// ------- Get Computing Parameters -------
	if (Na != 3) { mexErrMsgTxt("3 Parameters required! Radix [2,4,8] , TwidlleOption [0,1] & Shared memory usage [0,1]!\n\0"); }
	int FFTOpt     = static_cast<int>(Options[0]);
	int TwiddleOpt = static_cast<int>(Options[1]);
	int SharedOpt  = static_cast<int>(Options[2]);
	bool TwiddlePrecalculate = false;
	bool SharedMem = true;
	

	switch (FFTOpt) {
		case 2: {
			Base = 2;
			Dev.THREADS = dim3(1024, 1, 1);
			Dev.BLOCKS  = dim3(Nx / 1024 + 1, 1, 1);
			break;
		}
		case 4: {
			Base = 4;
			Dev.THREADS = dim3(512, 1, 1);
			Dev.BLOCKS = dim3(Nx / 512 + 1, 1, 1);
			break;
		}
		case 8: {
			Base = 8;
			Dev.THREADS = dim3(128, 1, 1);
			Dev.BLOCKS = dim3(Nx / 128 + 1, 1, 1);
			break;
		}
		default: {
			mexErrMsgTxt("UNSUPPORTED FFT RADIX! \n\0"); break;
		}
	}

	// --------------------------------
	// ------ Get Twiddle Option ------
	switch (TwiddleOpt) {
		case 0: {
			TwiddlePrecalculate = false; 
			break;
		}
		case 1: {
			TwiddlePrecalculate = true; 
			break;
		}
		default: {
			mexErrMsgTxt("! 0 --> Runtime Twiddle Calc | 1 --> Precalculate Twiddles! \n\0"); break;
		}
	}

	// ---------------------------------------
	// ------- Get Shared Memory Option ------
	switch (SharedOpt) {
		case 0: {
			SharedMem = false;
			break;
		}
		case 1: {
			SharedMem = true;
			break;
		}
		default: {
			mexErrMsgTxt("! 0 --> Global Memory FFT || 1 --> Shared ! \n\0");
			break;
		}
	}

	// ---------------------------------------------------------
	// ------- Compute Amount of Digits for DigitReversal ------
	int NumDigits = BaseLog<int>(Nx, Base);

	// --------------------------------------------
	// ------- Check AMount of CUDA Devices -------
	cudaGetDeviceCount(&Dev.NumDevices);
	if (Dev.NumDevices <= 0) {
		mexErrMsgTxt("No CUDA Supported Device Found! \n\0");
	}else {

		cudaGetDeviceProperties(&Dev.Prop, 0);
		std::string message = std::string("Device : ") + std::string(Dev.Prop.name) + std::string(" CC: ") + std::to_string(Dev.Prop.major) + "." + std::to_string(Dev.Prop.minor) + std::string("\n");
		mexPrintf(message.c_str());
	}

	// -----------------------------------
	// ------ Input Error Checking -------
	if (pow(Base, NumDigits) != Nx) { mexErrMsgTxt("Check Used Radix and Amount of Data!\n\0"); }

	// --------------------------------------
	// ------- Associate Matlab Output ------
	mxArray* M_Out_Data = plhs[0] = mxCreateNumericMatrix(1, Nx, mxSINGLE_CLASS, mxCOMPLEX);
	mxArray* M_Out_Inds = plhs[1] = mxCreateNumericMatrix(1, Nx, mxINT32_CLASS, mxREAL);
	mxComplexSingle* Mat_Data_Out = mxGetComplexSingles(M_Out_Data);
	mxInt32* Mat_Indexes_Out = mxGetInt32s(M_Out_Inds);
	
	// ------------------------------
	// ------- Create Events --------
	HANDLE_CUDA(cudaEventCreate(&Dev.Start));
	HANDLE_CUDA(cudaEventCreate(&Dev.Stop));
	HANDLE_CUDA(cudaEventRecord(Dev.Start));

	// -------------------------------------
	// ------ Allocate Device Memory -------
	HANDLE_CUDA(cudaMalloc((void**)&Dev.Data1, Nx * sizeof(cufftComplex)));
	HANDLE_CUDA(cudaMalloc((void**)&Dev.Data2, Nx * sizeof(cufftComplex)));
	HANDLE_CUDA(cudaMalloc((void**)&Dev.Twiddles, Nx * sizeof(cufftComplex)));
	HANDLE_CUDA(cudaMalloc((void**)&Dev.Indexes, Nx * sizeof(int)));

	// ----------------------------------
	// ------ Copy Data to Device -------
	HANDLE_CUDA(cudaMemcpy(Dev.Data1, MatData_In, Nx * sizeof(cufftComplex), cudaMemcpyHostToDevice));

	// ------------------------------------------
	// ------- Precompute Twiddle Factors -------
	if (TwiddlePrecalculate) {
		Kern_Compute_Twiddles << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Twiddles, Nx);
		HANDLE_CUDA(cudaGetLastError());
	}

	// ----------------------------------
	// ------- MAIN COMPUTING CORE ------
	int Distance = 0; 
	int Steps = BaseLog<int>(Nx, Base); 
	int Stage = 1;
	if (SharedMem) {
		switch (Base) {
			case RADIX_2: {
				int Shift;
				for (Stage = 1; Stage <= (Steps - 10); Stage++) { // 10 Steps Inside Shared Memory
					Shift = (Steps - Stage) * 1;
					Distance = (1 << Shift);
#ifdef DEBUG
					mexPrintf("Launching Global R2, Distance: %d Stage: %d Steps: %d Nx: %d \n", Distance, Stage, Steps, Nx);
#endif
					Kern_R2_Global << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
					HANDLE_CUDA(cudaDeviceSynchronize());
#endif
					HANDLE_CUDA(cudaGetLastError());
				}

				// ---------------------------------------
				// ------- FINISH IN SHARED MEMORY -------
				// NOTE: Use 512 Threads/Block!
				Shift = (Steps - Stage) * 1;
				Distance = (1 << Shift);
#ifdef DEBUG
				mexPrintf("Launching Shared R2, Distance: %d Stage: %d Steps: %d Nx: %d \n", Distance, Stage, Steps, Nx);
#endif
				Kern_R2_Shared << <dim3(Nx / 512 + 1, 1, 1), dim3(512, 1, 1) >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
				HANDLE_CUDA(cudaDeviceSynchronize());
#endif
				HANDLE_CUDA(cudaGetLastError());
				break;
			}

			case RADIX_4: {
				int Shift;
				for (Stage = 1; Stage <= (Steps - 5); Stage++) { // 5 Steps Inside Shared Memory
					Shift = (Steps - Stage) * 2;
					Distance = (1 << Shift);
#ifdef DEBUG
					mexPrintf("Launching Global R4, Distance: %d Stage: %d Steps: %d Nx: %d \n", Distance, Stage, Steps, Nx);
#endif
					Kern_R4_Global << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
					HANDLE_CUDA(cudaDeviceSynchronize());
#endif
					HANDLE_CUDA(cudaGetLastError());
				}

				// ---------------------------------------
				// ------- FINISH IN SHARED MEMORY -------
				// NOTE: Use 256 Threads/Block!
				Shift = (Steps - Stage) * 2;
				Distance = (1 << Shift);
#ifdef DEBUG
				mexPrintf("Launching Shared R4, Distance: %d Stage: %d Steps: %d Nx: %d \n", Distance, Stage, Steps, Nx);
#endif
				Kern_R4_Shared << <dim3(Nx / 256 + 1, 1, 1), dim3(256, 1, 1) >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
				HANDLE_CUDA(cudaDeviceSynchronize());
#endif
				HANDLE_CUDA(cudaGetLastError());
				break;
			}

			case RADIX_8: {
				int Shift;
				for (Stage = 1; Stage <= (Steps - 3); Stage++) { // 3 Steps Inside Shared Memory
					Shift = (Steps - Stage) * 3;
					Distance = (1 << Shift);
#ifdef DEBUG
					mexPrintf("Launching Global R8, Distance: %d Stage: %d Steps: %d Nx: %d \n", Distance, Stage, Steps, Nx);
#endif
					Kern_R8_Global << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
					HANDLE_CUDA(cudaDeviceSynchronize());
#endif
					HANDLE_CUDA(cudaGetLastError());
				}

				// ---------------------------------------
				// ------- FINISH IN SHARED MEMORY -------
				// NOTE: Use 64 Threads/Block!
				Shift = (Steps - Stage) * 3;
				Distance = (1 << Shift);
#ifdef DEBUG
				mexPrintf("Launching Shared R8, Distance: %d Stage: %d Steps: %d Nx: %d \n", Distance,Stage,Steps,Nx);
#endif
				Kern_R8_Shared << <dim3(Nx / 64 + 1, 1, 1), dim3(64, 1, 1) >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
				HANDLE_CUDA(cudaDeviceSynchronize());
#endif
				HANDLE_CUDA(cudaGetLastError());
				break;
			}
		}

	}else {

		// --------------------------------------
		// ------- Use Global Memory Only -------
		for (int Stage = 1; Stage <= Steps; Stage++) {
			switch (Base) {
				case RADIX_2: {
					int Shift = (Steps - Stage) * 1;
					Distance = (1 << Shift);
					Kern_R2_Global << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
					HANDLE_CUDA(cudaDeviceSynchronize());
#endif
					HANDLE_CUDA(cudaGetLastError());
					break;
				}
				case RADIX_4: {
					int Shift = (Steps - Stage) * 2;
					Distance = (1 << Shift);
					Kern_R4_Global << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
					HANDLE_CUDA(cudaDeviceSynchronize());
#endif
					HANDLE_CUDA(cudaGetLastError());
					break;
				}
				case RADIX_8: {
					int Shift = (Steps - Stage) * 3;
					Distance = (1 << Shift);
					Kern_R8_Global << <Dev.BLOCKS, Dev.THREADS >> > (Dev.Data1, Dev.Twiddles, Distance, Stage, Steps, Nx, TwiddlePrecalculate);
#ifdef DEBUG
					HANDLE_CUDA(cudaDeviceSynchronize());
#endif
					HANDLE_CUDA(cudaGetLastError());
					break;
				}
				default: {
					std::string Message = std::string("Error Unsupported Radix!\n");
					mexErrMsgTxt(Message.c_str());
					break;
				}
			}
		}
	}

	// ------------------------------------
	// ------- Digit Reversal (DIF) -------
	dim3 TH_DIGITS(256,1,1);
	dim3 BL_DIGITS(Nx/256 + 1,1,1);
	Kern_DigitReversal << <BL_DIGITS, TH_DIGITS >> > (Dev.Data1, Dev.Data2, Dev.Indexes, Nx, Base, NumDigits);
#ifdef DEBUG
	HANDLE_CUDA(cudaDeviceSynchronize());
#endif
	HANDLE_CUDA(cudaGetLastError());

	// ------------------------------------------
	// -------- Copy Results back to Host -------
	HANDLE_CUDA(cudaDeviceSynchronize());
	HANDLE_CUDA(cudaMemcpy(Mat_Data_Out, Dev.Data2, Nx * sizeof(cufftComplex), cudaMemcpyDeviceToHost));
	HANDLE_CUDA(cudaMemcpy(Mat_Indexes_Out, Dev.Indexes, Nx * sizeof(int), cudaMemcpyDeviceToHost));

	// ------------------------------------
	// ------- Compute Elapsed Time -------
	float ElapsedTime = 0.0f;
	HANDLE_CUDA(cudaEventRecord(Dev.Stop));
	HANDLE_CUDA(cudaEventSynchronize(Dev.Stop));
	HANDLE_CUDA(cudaEventElapsedTime(&ElapsedTime, Dev.Start, Dev.Stop));
	std::string Message = std::string("GPUFFT ElapsedTime: ") + std::to_string(ElapsedTime) + std::string(" ms\n\0");
	mexPrintf(Message.c_str());

	// ----------------------------
	// ------- Clean Memory -------
	HANDLE_CUDA(cudaFree(Dev.Data1));
	HANDLE_CUDA(cudaFree(Dev.Data2));
	HANDLE_CUDA(cudaFree(Dev.Twiddles));
	HANDLE_CUDA(cudaFree(Dev.Indexes));
	HANDLE_CUDA(cudaEventDestroy(Dev.Start));
	HANDLE_CUDA(cudaEventDestroy(Dev.Stop));

}
