/**********************************
 * blog.iriscores.com 2016/2022
 * Author: vojta.ters@iriscores.com
 **********************************/

// ------------------------------
// -------- Cuda Includes -------
#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <cufft.h>
#include <math_constants.h>

// -----------------------------
// ------- C++ Includes --------
#include <stdio.h>
#include <string>

struct R2Complex {
	cufftComplex A;
	cufftComplex B;
};

struct R4Complex {
	cufftComplex A;
	cufftComplex B;
	cufftComplex C;
	cufftComplex D;
};

struct R8Complex {
	cufftComplex X1;
	cufftComplex X2;
	cufftComplex X3;
	cufftComplex X4;
	cufftComplex X5;
	cufftComplex X6;
	cufftComplex X7;
	cufftComplex X8;
};


/****************************************************
 * @brief: Custom Complex Number Operator Overloading
 ****************************************************/
__device__ cufftComplex operator+(cufftComplex A, cufftComplex B) {
	cufftComplex Cop;
	Cop.x = A.x + B.x;
	Cop.y = A.y + B.y;
	return Cop;
}
__device__ cufftComplex operator-(cufftComplex A, cufftComplex B) {
	cufftComplex Cop;
	Cop.x = A.x - B.x;
	Cop.y = A.y - B.y;
	return Cop;
}
__device__ cufftComplex operator-(cufftComplex A) {
	A.x = -A.x;
	A.y = -A.y;
	return A;
}
__device__ cufftComplex operator*(cufftComplex A, cufftComplex B) {
	cufftComplex Cop;
	Cop.x = A.x * B.x - A.y * B.y;
	Cop.y = A.y * B.x + A.x * B.y;
	return Cop;
}
__device__ cufftComplex operator*(cufftComplex A, cufftReal B) {
	cufftComplex Cop;
	Cop.x = A.x * B;
	Cop.y = A.y * B;
	return Cop;
}


/**********************
 * @Brief: Abs(Complex)
 **********************/
__device__ cufftReal CA(cufftComplex A) {
	return sqrtf(A.x * A.x + A.y * A.y);
}

/****************************
 * @brief: Complex(Real,Real)
 ****************************/
__device__ cufftComplex CM(cufftReal A, cufftReal B) {
	cufftComplex Cop;
	Cop.x = A; Cop.y = B;
	return Cop;
}

/**********************
 *@brief: conj(Complex)
 * NOTE: Complex Conjugation
 **********************/
__device__ cufftComplex CC(cufftComplex A) {
	cufftComplex Cop = A;
	Cop.y = Cop.y * -1.0f;
	return Cop;
}

/************************************
 *@brief: Complex(A) * 1i
 * NOTE: Multiply with Imaginary Unit
 ************************************/
__device__ cufftComplex CI(cufftComplex A) {
	cufftComplex Cop;
	Cop.x = -A.y;
	Cop.y = A.x;
	return Cop;
}

/***************************************************************
 * @brief: Decomposes Index into Base Radix Digits
 *  WARNING: powf causes problems with precision! Avoid if able!
 ***************************************************************/
__device__ void Dec2Base(int*S_Digits,int Num, int Base, int NumDigits) {

	int Index = Num;
	for (int i = 0; i < NumDigits; i++) {
		int Power = 1;
		switch (Base) {
			case 2: {
				// 2^(NumDigits - i - 1)
				int Shift = (NumDigits - i - 1) * 1;
				Power = (1 << Shift);
				break;
			}
			case 4: {
				// 4^(NumDigits - i - 1)
				int Shift = (NumDigits - i - 1) * 2;
				Power = (1 << Shift);
				break;
			}
			case 8: {
				// 8^(NumDigits - i - 1)
				int Shift = (NumDigits - i - 1) * 3;
				Power = (1 << Shift);
				break;
			}
		}
		int Tmp = Index / Power;					// Integer Division
		Index -= Tmp * Power;						// Substract
		S_Digits[i] = Tmp;							// Save Digits
	}
}

/*****************************
 *@brief: Reverses Base Digits
 *****************************/
__device__ void DigitRewOrder( int* S_Digits, int NumDigits) {

	for (int i = 0; i < NumDigits / 2; i++) {
		int Digit_Left = S_Digits[i];
		int Digit_Right = S_Digits[NumDigits - i - 1];

		S_Digits[i] = Digit_Right;
		S_Digits[NumDigits - i - 1] = Digit_Left;
	}
}

/**************************************************************
 * @brief: Converts Digits back to a real Index
 * WARNING: powf causes problems with precision! Avoid if able!
 **************************************************************/
__device__ int Base2Dec(int* S_Digits, int Base, int NumDigits) {
	int ret = 0;
	for (int i = 0; i < NumDigits; i++) {
		int Power = 1;
		switch (Base) {
			case 2: {
				// 2^(NumDigits - i - 1)
				int Shift = (NumDigits - i - 1) * 1;
				Power = (1 << Shift);
				break;
			}
			case 4: {
				// 4^(NumDigits - i - 1)
				int Shift = (NumDigits - i - 1) * 2;
				Power = (1 << Shift);
				break;
			}
			case 8: {
				// 8^(NumDigits - i - 1)
				int Shift = (NumDigits - i - 1) * 3;
				Power = (1 << Shift);
				break;
			}
		}
		ret += S_Digits[i] * Power;
	}
	return ret;
}

/********************************************
 * @brief: Computes the Radix-2 FFT Butterfly
 *******************************************/
__device__ R2Complex Radix2Butterfly(cufftComplex A, cufftComplex B, cufftComplex T1, cufftComplex T2) {
	R2Complex Rout;
	Rout.A = (A + B) * T1;
	Rout.B = (A - B) * T2;
	return Rout;
}

/*******************************************
 *@brief: Computes the Radix-4 FFT Butterfly
 ******************************************/
__device__ R4Complex Radix4Butterfly(cufftComplex A, cufftComplex B, cufftComplex C, cufftComplex D, cufftComplex T1, cufftComplex T2, cufftComplex T3, cufftComplex T4) {
	R4Complex Rout;
	Rout.A = (A + B + C + D) * T1;
	Rout.C = (A - B + C - D) * T3;
	Rout.B = (A - CI(B) - C + CI(D)) * T2;
	Rout.D = (A + CI(B) - C - CI(D)) * T4;
	return Rout;
}

/*******************************************
 *@brief: Computes the Radix-8 FFT Butterfly
 ******************************************/
__device__ R8Complex Radix8Butterfly(cufftComplex A, cufftComplex B, cufftComplex C, cufftComplex D, cufftComplex E, cufftComplex F, cufftComplex G, cufftComplex H, cufftComplex T1, cufftComplex T2, cufftComplex T3, cufftComplex T4, cufftComplex T5, cufftComplex T6, cufftComplex T7, cufftComplex T8) {
	R8Complex Rout;
	Rout.X1 = (A + B + C + D + E + F + G + H) * T1;
	Rout.X5 = (A - B + C - D + E - F + G - H) * T5;
	Rout.X3 = (A - CI(B) - C + CI(D) + E - CI(F) - G + CI(H)) * T3;
	Rout.X7 = (A + CI(B) - C - CI(D) + E + CI(F) - G - CI(H)) * T7;

	Rout.X2 = (A + CM(+CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * B - CI(C) + CM(-CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * D - E + CM(-CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * F + CI(G) + CM(+CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * H) * T2;
	Rout.X4 = (A + CM(-CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * B + CI(C) + CM(+CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * D - E + CM(+CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * F - CI(G) + CM(-CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * H) * T4;
	Rout.X6 = (A + CM(-CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * B - CI(C) + CM(+CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * D - E + CM(+CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * F + CI(G) + CM(-CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * H) * T6;
	Rout.X8 = (A + CM(+CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * B + CI(C) + CM(-CUDART_SQRT_HALF_F, +CUDART_SQRT_HALF_F) * D - E + CM(-CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * F - CI(G) + CM(+CUDART_SQRT_HALF_F, -CUDART_SQRT_HALF_F) * H) * T8;
	return Rout;
}

/*****************************************************
 *@brief: Does DigitReversal for a specified FFT Radix
 * NOTE: Requires 256 Threds per Block!
 ****************************************************/
__global__ void Kern_DigitReversal(cufftComplex* InData, cufftComplex* OutData,int *Indexes, int Nx, int Base, int NumDigits) {
	int tx = threadIdx.x;
	int tix = threadIdx.x + blockIdx.x * blockDim.x;

	// -------------------------------------------------------------------
	// -------- Have Each Thread assigned a piece of Shared Memory -------
	// Maximum Amount of Digits: 32
	__shared__ int SB_Digits[256 * 32];
	int* S_Digits = &SB_Digits[tx * 32];

	if (tix < Nx) {

		Dec2Base(S_Digits, tix, Base, NumDigits);	      // TO BASE DIGITS
		DigitRewOrder(S_Digits, NumDigits);               // REWRITE ORDER
		int tidOut = Base2Dec(S_Digits, Base, NumDigits); // BACK TO DECIMAL

		// -----------------------------------
		// ------- Save A copy of Index ------
		Indexes[tix] = tidOut;

		// ------------------------------------------------------
		// ------- Do the DigitReversal in Global memory --------
		OutData[tidOut] = InData[tix];
	}
}

/*************************************
 *@brief: Precomputes Twiddles for FFT
 *************************************/
__global__ void Kern_Compute_Twiddles(cufftComplex* Twiddles, int N) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	if (tix < N) {
		Twiddles[tix].x = cosf(-2.0f * CUDART_PI_F * (cufftReal)tix / (cufftReal)N);
		Twiddles[tix].y = sinf(-2.0f * CUDART_PI_F * (cufftReal)tix / (cufftReal)N);
	}
}

/*************************************************************
 *  @brief: Computes the Radix2 FFT Stage within Global Memory
 ************************************************************/
__global__ void Kern_R2_Global(cufftComplex* Data, const cufftComplex* Twiddles, int Distance, int Stage, int Steps, int Nx, bool TwiddlePrecalculate) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	if (tix < Nx / 2) {

		// ---------------------------------------
		// ------- Compute Butterfly Index -------
		int Ind = fmodf(tix, Distance) + (tix / Distance) * Distance * 2;

		// -----------------------------------------------
		// ------- Load or Compute Twiddle Factors -------
		cufftComplex T1, T2;
		if (TwiddlePrecalculate) {
			int Shift = (Stage - 1) * 1;
			int Power = (1 << Shift);
			int TWLI = Power * fmodf(tix, Distance);
			T1 = Twiddles[0 * TWLI];
			T2 = Twiddles[1 * TWLI];
		}else {
			if (Stage == Steps) {
				T1.x = 1.0f; T1.y = 0.0f;
				T2.x = 1.0f; T2.y = 0.0f;
			}else {
				T1.x = cosf(0.0f / Nx * powf(2.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T1.y = sinf(0.0f / Nx * powf(2.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T2.x = cosf(1.0f / Nx * powf(2.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T2.y = sinf(1.0f / Nx * powf(2.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
			}
		}

		// -----------------------------------
		// ------- Calculate Butterfly -------
		R2Complex BTF = Radix2Butterfly(Data[Ind], Data[Ind + Distance], T1, T2);

		// ----------------------------------------------------
		// ------- Save Computed Results Back to Global -------
		Data[Ind + 0 * Distance] = BTF.A;
		Data[Ind + 1 * Distance] = BTF.B;
	}
}

/***********************************************************
 *@brief: Computes the Radix2 FFT Stage Within Shared Memory
 * NOTE: Requires 512 Threads/Block
 ***********************************************************/
__global__ void Kern_R2_Shared(cufftComplex* Data, const cufftComplex* Twiddles, int Distance, int Stage, int Steps, int Nx, bool TwiddlePrecalculate) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tx = threadIdx.x;
	if (tix < Nx / 2) {

		// -----------------------------------------
		// ------- Load from Global And Sync -------
		cufftComplex T1, T2; int StageId;
		__shared__ cufftComplex SData[1024];
		SData[tx * 2 + 0] = Data[tix * 2 + 0];
		SData[tx * 2 + 1] = Data[tix * 2 + 1];
		__syncthreads();

		// ----------------------------------------
		// ------- Shared Memory Stage Loop -------
		for (StageId = Stage; StageId <= Steps; StageId++) {

			// ---------------------------------------
			// ------- Compute Butterfly Index -------
			int Ind = fmodf(tx, Distance) + (tx / Distance) * Distance * 2;

			// -----------------------------------------------
			// ------- Load or Compute Twiddle Factors -------
			if (TwiddlePrecalculate) {
				int Shift = (StageId - 1) * 1;
				int Power = (1 << Shift);
				int TWLI = Power * fmodf(tix, Distance);
				T1 = Twiddles[0 * TWLI];
				T2 = Twiddles[1 * TWLI];
			}else {
				if (StageId == Steps) {
					T1.x = 1.0f; T1.y = 0.0f;
					T2.x = 1.0f; T2.y = 0.0f;
				}else {
					T1.x = cosf(0.0f / Nx * powf(2.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T1.y = sinf(0.0f / Nx * powf(2.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T2.x = cosf(1.0f / Nx * powf(2.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T2.y = sinf(1.0f / Nx * powf(2.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				}
			}

			// -----------------------------------
			// ------- Calculate Butterfly -------
			R2Complex BTF = Radix2Butterfly(SData[Ind], SData[Ind + Distance], T1, T2);
			SData[Ind + 0 * Distance] = BTF.A;
			SData[Ind + 1 * Distance] = BTF.B;

			// -----------------------------------------
			// ------- Update Butterfly Distance -------
			Distance = Distance / 2;

			// ----------------------------------------
			// ------- Always Sync For Next Step ------
			__syncthreads();
		}

		// ----------------------------------------------------
		// ------- Save Computed Results Back to Global -------
		Data[tix * 2 + 0] = SData[tx * 2 + 0];
		Data[tix * 2 + 1] = SData[tx * 2 + 1];
	}
}

/***********************************************************
 * @brief Computes the Radix4 FFT Stage within Global Memory
 ***********************************************************/
__global__ void Kern_R4_Global(cufftComplex* Data, const cufftComplex* Twiddles, int Distance, int Stage, int Steps, int Nx, bool TwiddlePrecalculate) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	if (tix < Nx / 4) {

		// ---------------------------------------
		// ------- Compute Butterfly Index -------
		int Ind = fmodf(tix, Distance) + (tix / Distance) * Distance * 4;

		// -----------------------------------------------
		// ------- Load or Compute Twiddle Factors -------
		cufftComplex T1, T2, T3, T4;
		if (TwiddlePrecalculate) {
			int Shift = (Stage - 1) * 2;
			int Power = (1 << Shift);
			int TWLI = Power * fmodf(tix, Distance);
			T1 = Twiddles[0 * TWLI];
			T2 = Twiddles[1 * TWLI];
			T3 = Twiddles[2 * TWLI];
			T4 = Twiddles[3 * TWLI];

		}else {
			if (Stage == Steps) {
				T1.x = 1.0f; T1.y = 0.0f;
				T2.x = 1.0f; T2.y = 0.0f;
				T3.x = 1.0f; T3.y = 0.0f;
				T4.x = 1.0f; T4.y = 0.0f;
			}else {
				T1.x = cosf(0.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T1.y = sinf(0.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T2.x = cosf(1.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T2.y = sinf(1.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T3.x = cosf(2.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T3.y = sinf(2.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T4.x = cosf(3.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T4.y = sinf(3.0f / Nx * powf(4.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
			}
		}

		// -----------------------------------
		// ------- Calculate Butterfly -------
		R4Complex BTF = Radix4Butterfly(Data[Ind], Data[Ind + Distance], Data[Ind + 2 * Distance], Data[Ind + 3 * Distance], T1, T2, T3, T4);

		// ----------------------------------------------------
		// ------- Save Computed Results Back to Global -------
		Data[Ind + 0 * Distance] = BTF.A;
		Data[Ind + 1 * Distance] = BTF.B;
		Data[Ind + 2 * Distance] = BTF.C;
		Data[Ind + 3 * Distance] = BTF.D;
	}
}

/***********************************************************
 * @brief Computes the Radix4 FFT Stage within Shared Memory
 * NOTE: Requires 256 Threads/Block
 ***********************************************************/
__global__ void Kern_R4_Shared(cufftComplex* Data, const cufftComplex* Twiddles, int Distance, int Stage, int Steps, int Nx, bool TwiddlePrecalculate) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tx = threadIdx.x;
	if (tix < Nx / 4) {

		// -----------------------------------------
		// ------- Load from Global And Sync -------
		cufftComplex T1, T2, T3, T4; int StageId;
		__shared__ cufftComplex SData[1024];
		SData[tx * 4 + 0] = Data[tix * 4 + 0];
		SData[tx * 4 + 1] = Data[tix * 4 + 1];
		SData[tx * 4 + 2] = Data[tix * 4 + 2];
		SData[tx * 4 + 3] = Data[tix * 4 + 3];
		__syncthreads();

		// ----------------------------------------
		// ------- Shared Memory Stage Loop -------
		for (StageId = Stage; StageId <= Steps; StageId++) {

			// ---------------------------------------
			// ------- Compute Butterfly Index -------
			int Ind = fmodf(tx, Distance) + (tx / Distance) * Distance * 4;

			// -----------------------------------------------
			// ------- Load or Compute Twiddle Factors -------
			if (TwiddlePrecalculate) {
				int Shift = (StageId - 1) * 2;
				int Power = (1 << Shift);
				int TWLI = Power * fmodf(tix, Distance);
				T1 = Twiddles[0 * TWLI];
				T2 = Twiddles[1 * TWLI];
				T3 = Twiddles[2 * TWLI];
				T4 = Twiddles[3 * TWLI];
			}else {
				if (Stage == Steps) {
					T1.x = 1.0f; T1.y = 0.0f;
					T2.x = 1.0f; T2.y = 0.0f;
					T3.x = 1.0f; T3.y = 0.0f;
					T4.x = 1.0f; T4.y = 0.0f;
				}else {
					T1.x = cosf(0.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T1.y = sinf(0.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T2.x = cosf(1.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T2.y = sinf(1.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T3.x = cosf(2.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T3.y = sinf(2.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T4.x = cosf(3.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T4.y = sinf(3.0f / Nx * powf(4.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				}

			}

			// -----------------------------------
			// ------- Calculate Butterfly -------
			R4Complex BTF = Radix4Butterfly(SData[Ind], SData[Ind + Distance], SData[Ind + 2 * Distance], SData[Ind + 3 * Distance], T1, T2, T3, T4);
			SData[Ind + 0 * Distance] = BTF.A;
			SData[Ind + 1 * Distance] = BTF.B;
			SData[Ind + 2 * Distance] = BTF.C;
			SData[Ind + 3 * Distance] = BTF.D;

			// -----------------------------------------
			// ------- Update Butterfly Distance -------
			Distance = Distance / 4;

			// ----------------------------------------
			// ------- Always Sync For Next Step ------
			__syncthreads();

		}

		// ----------------------------------------------------
		// ------- Save Computed Results Back to Global -------
		Data[tix * 4 + 0] = SData[tx * 4 + 0];
		Data[tix * 4 + 1] = SData[tx * 4 + 1];
		Data[tix * 4 + 2] = SData[tx * 4 + 2];
		Data[tix * 4 + 3] = SData[tx * 4 + 3];

	}
}

/************************************************************
 * @brief: Computes the Radix8 FFT Stage within Global Memory
 ************************************************************/
__global__ void Kern_R8_Global(cufftComplex* Data, const cufftComplex* Twiddles, int Distance, int Stage, int Steps, int Nx, bool TwiddlePrecalculate) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	if (tix < Nx / 8) {

		// ---------------------------------------
		// ------- Compute Butterfly Index -------
		int Ind = fmodf(tix, Distance) + (tix / Distance) * Distance * 8;

		// -----------------------------------------------
		// ------- Load or Compute Twiddle Factors -------
		cufftComplex T1, T2, T3, T4, T5, T6, T7, T8;
		if (TwiddlePrecalculate) {
			int Shift = (Stage - 1) * 3;
			int Power = (1 << Shift);
			int TWLI = Power * fmodf(tix, Distance);
			T1 = Twiddles[0 * TWLI];
			T2 = Twiddles[1 * TWLI];
			T3 = Twiddles[2 * TWLI];
			T4 = Twiddles[3 * TWLI];
			T5 = Twiddles[4 * TWLI];
			T6 = Twiddles[5 * TWLI];
			T7 = Twiddles[6 * TWLI];
			T8 = Twiddles[7 * TWLI];
		}else {
			if (Stage == Steps) {
				T1.x = 1.0f; T1.y = 0.0f;
				T2.x = 1.0f; T2.y = 0.0f;
				T3.x = 1.0f; T3.y = 0.0f;
				T4.x = 1.0f; T4.y = 0.0f;
				T5.x = 1.0f; T5.y = 0.0f;
				T6.x = 1.0f; T6.y = 0.0f;
				T7.x = 1.0f; T7.y = 0.0f;
				T8.x = 1.0f; T8.y = 0.0f;
			}else {
				T1.x = cosf(0.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T1.y = sinf(0.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T2.x = cosf(1.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T2.y = sinf(1.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T3.x = cosf(2.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T3.y = sinf(2.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T4.x = cosf(3.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T4.y = sinf(3.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T5.x = cosf(4.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T5.y = sinf(4.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T6.x = cosf(5.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T6.y = sinf(5.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T7.x = cosf(6.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T7.y = sinf(6.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T8.x = cosf(7.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				T8.y = sinf(7.0f / Nx * powf(8.0f, Stage - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
			}
		}

		// -----------------------------------
		// ------- Calculate Butterfly -------
		R8Complex BTF = Radix8Butterfly(Data[Ind], Data[Ind + Distance], Data[Ind + 2 * Distance], Data[Ind + 3 * Distance], Data[Ind + 4 * Distance], Data[Ind + 5 * Distance], Data[Ind + 6 * Distance], Data[Ind + 7 * Distance], T1, T2, T3, T4, T5, T6, T7, T8);

		// ----------------------------------------------------
		// ------- Save Computed Results Back to Global -------
		Data[Ind + 0 * Distance] = BTF.X1;
		Data[Ind + 1 * Distance] = BTF.X2;
		Data[Ind + 2 * Distance] = BTF.X3;
		Data[Ind + 3 * Distance] = BTF.X4;
		Data[Ind + 4 * Distance] = BTF.X5;
		Data[Ind + 5 * Distance] = BTF.X6;
		Data[Ind + 6 * Distance] = BTF.X7;
		Data[Ind + 7 * Distance] = BTF.X8;
	}
}

/************************************************************
 * @brief: Computes the Radix8 FFT Stage Within Shared Memory
 * NOTE: Requires 64 Threads/Block
 ***********************************************************/
__global__ void Kern_R8_Shared(cufftComplex* Data, const cufftComplex* Twiddles, int Distance, int Stage, int Steps, int Nx, bool TwiddlePrecalculate) {
	int tix = threadIdx.x + blockIdx.x * blockDim.x;
	int tx = threadIdx.x;
	if (tix < Nx / 8) {


		// -----------------------------------------
		// ------- Load from Global And Sync -------
		cufftComplex T1, T2, T3, T4, T5, T6, T7, T8; int StageId;
		__shared__ cufftComplex SData[512];
		SData[tx * 8 + 0] = Data[tix * 8 + 0];
		SData[tx * 8 + 1] = Data[tix * 8 + 1];
		SData[tx * 8 + 2] = Data[tix * 8 + 2];
		SData[tx * 8 + 3] = Data[tix * 8 + 3];
		SData[tx * 8 + 4] = Data[tix * 8 + 4];
		SData[tx * 8 + 5] = Data[tix * 8 + 5];
		SData[tx * 8 + 6] = Data[tix * 8 + 6];
		SData[tx * 8 + 7] = Data[tix * 8 + 7];
		__syncthreads();

		// ----------------------------------------
		// ------- Shared Memory Stage Loop -------
		for (StageId = Stage; StageId <= Steps; StageId++) {

			// ---------------------------------------
			// ------- Compute Butterfly Index -------
			int Ind = fmodf(tx, Distance) + (tx / Distance) * Distance * 8;

			// -----------------------------------------------
			// ------- Load or Compute Twiddle Factors -------
			if (TwiddlePrecalculate) {
				int Shift = (StageId - 1) * 3;
				int Power = (1 << Shift);
				int TWLI = Power * fmodf(tix, Distance);
				T1 = Twiddles[0 * TWLI];
				T2 = Twiddles[1 * TWLI];
				T3 = Twiddles[2 * TWLI];
				T4 = Twiddles[3 * TWLI];
				T5 = Twiddles[4 * TWLI];
				T6 = Twiddles[5 * TWLI];
				T7 = Twiddles[6 * TWLI];
				T8 = Twiddles[7 * TWLI];

			}else {
				if (Stage == Steps) {
					T1.x = 1.0f; T1.y = 0.0f;
					T2.x = 1.0f; T2.y = 0.0f;
					T3.x = 1.0f; T3.y = 0.0f;
					T4.x = 1.0f; T4.y = 0.0f;
					T5.x = 1.0f; T5.y = 0.0f;
					T6.x = 1.0f; T6.y = 0.0f;
					T7.x = 1.0f; T7.y = 0.0f;
					T8.x = 1.0f; T8.y = 0.0f;
				}else {
					T1.x = cosf(0.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T1.y = sinf(0.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T2.x = cosf(1.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T2.y = sinf(1.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T3.x = cosf(2.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T3.y = sinf(2.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T4.x = cosf(3.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T4.y = sinf(3.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T5.x = cosf(4.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T5.y = sinf(4.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T6.x = cosf(5.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T6.y = sinf(5.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T7.x = cosf(6.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T7.y = sinf(6.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T8.x = cosf(7.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
					T8.y = sinf(7.0f / Nx * powf(8.0f, StageId - 1.0f) * fmodf(tix, Distance) * -2.0f * CUDART_PI_F);
				}

			}

			// -----------------------------------
			// ------- Calculate Butterfly -------
			R8Complex BTF = Radix8Butterfly(SData[Ind], SData[Ind + Distance], SData[Ind + 2 * Distance], SData[Ind + 3 * Distance], SData[Ind + 4 * Distance], SData[Ind + 5 * Distance], SData[Ind + 6 * Distance], SData[Ind + 7 * Distance], T1, T2, T3, T4, T5, T6, T7, T8);
			SData[Ind + 0 * Distance] = BTF.X1;
			SData[Ind + 1 * Distance] = BTF.X2;
			SData[Ind + 2 * Distance] = BTF.X3;
			SData[Ind + 3 * Distance] = BTF.X4;
			SData[Ind + 4 * Distance] = BTF.X5;
			SData[Ind + 5 * Distance] = BTF.X6;
			SData[Ind + 6 * Distance] = BTF.X7;
			SData[Ind + 7 * Distance] = BTF.X8;


			// -----------------------------------------
			// ------- Update Butterfly Distance -------
			Distance = Distance / 8;

			// ----------------------------------------
			// ------- Always Sync For Next Step ------
			__syncthreads();

		}

		// ----------------------------------------------------
		// ------- Save Computed Results Back to Global -------
		Data[tix * 8 + 0] = SData[tx * 8 + 0];
		Data[tix * 8 + 1] = SData[tx * 8 + 1];
		Data[tix * 8 + 2] = SData[tx * 8 + 2];
		Data[tix * 8 + 3] = SData[tx * 8 + 3];
		Data[tix * 8 + 4] = SData[tx * 8 + 4];
		Data[tix * 8 + 5] = SData[tx * 8 + 5];
		Data[tix * 8 + 6] = SData[tx * 8 + 6];
		Data[tix * 8 + 7] = SData[tx * 8 + 7];

	}

}
